<?php

declare(strict_types = 1);

namespace Drupal\meta_entity;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\meta_entity\Entity\MetaEntityInterface;

/**
 * Provides a default implementation for 'meta_entity.repository' service.
 */
class MetaEntityRepository implements MetaEntityRepositoryInterface {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The cache chained backend.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  protected $cache;

  /**
   * Constructs a new service instance.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager service.
   * @param \Drupal\Core\Cache\CacheBackendInterface $cache
   *   The cache chained backend.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, CacheBackendInterface $cache) {
    $this->entityTypeManager = $entity_type_manager;
    $this->cache = $cache;
  }

  /**
   * {@inheritdoc}
   */
  public function getMetaEntityForEntity(ContentEntityInterface $entity, string $meta_entity_type_id): ?MetaEntityInterface {
    $cid = "meta:{$meta_entity_type_id}:{$entity->getEntityTypeId()}:{$entity->id()}";

    // Try to retrieve from cache.
    if ($cache = $this->cache->get($cid)) {
      return $cache->data;
    }

    $storage = $this->entityTypeManager->getStorage('meta_entity');
    $ids = $storage->getQuery()
      ->condition('type', $meta_entity_type_id)
      ->condition('target.target_type', $entity->getEntityTypeId())
      ->condition('target.target_id', $entity->id())
      ->execute();
    $id = reset($ids);
    /** @var \Drupal\meta_entity\Entity\MetaEntityInterface $meta_entity */
    $meta_entity = $id ? $storage->load($id) : NULL;

    // Cache the results.
    $cache_tags = $this->entityTypeManager->getDefinition('meta_entity')->getListCacheTags();
    if ($meta_entity) {
      Cache::mergeTags($cache_tags, $meta_entity->getCacheTags());
    }
    $this->cache->set($cid, $meta_entity, Cache::PERMANENT, $cache_tags);

    return $meta_entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getMetaEntitiesForEntity(ContentEntityInterface $entity): array {
    $cid = "meta:{$entity->getEntityTypeId()}:{$entity->id()}";

    // Try to retrieve from cache.
    if ($cache = $this->cache->get($cid)) {
      return $cache->data;
    }

    $storage = $this->entityTypeManager->getStorage('meta_entity');
    $ids = $storage->getQuery()
      ->condition('target.target_type', $entity->getEntityTypeId())
      ->condition('target.target_id', $entity->id())
      // Make the order predictable.
      ->sort('type')
      ->execute();
    $meta_entities = $ids ? $storage->loadMultiple($ids) : [];

    // Cache the results.
    $cache_tags = $this->entityTypeManager->getDefinition('meta_entity')->getListCacheTags();
    array_walk($meta_entities, function (MetaEntityInterface $meta_entity) use (&$cache_tags): void {
      $cache_tags = Cache::mergeTags($cache_tags, $meta_entity->getCacheTags());
    });
    $this->cache->set($cid, $meta_entities, Cache::PERMANENT, $cache_tags);

    return $meta_entities;
  }

  /**
   * {@inheritdoc}
   */
  public function getMetaEntityTypesForBundle(string $entity_type_id, ?string $bundle = NULL): array {
    $bundle = $bundle ?: $entity_type_id;
    $cid = "types:{$entity_type_id}:{$bundle}";

    if ($cache = $this->cache->get($cid)) {
      return $cache->data;
    }

    $meta_types = [];
    $cache_tags = $this->entityTypeManager->getDefinition('meta_entity_type')->getListCacheTags();
    $storage = $this->entityTypeManager->getStorage('meta_entity_type');
    foreach ($storage->loadMultiple() as $id => $type) {
      $cache_tags = Cache::mergeTags($cache_tags, $type->getCacheTags());
      $mapping = $type->get('mapping');
      if (array_key_exists($entity_type_id, $mapping)) {
        if (array_key_exists($bundle, $mapping[$entity_type_id])) {
          $meta_types[] = $id;
        }
      }
    }

    if ($meta_types) {
      sort($meta_types);
    }

    // Cache the results.
    $this->cache->set($cid, $meta_types, Cache::PERMANENT, $cache_tags);

    return $meta_types;
  }

  /**
   * {@inheritdoc}
   */
  public function getReverseReferenceFieldNames(string $entity_type_id, string $bundle): array {
    $cid = "fields:{$entity_type_id}:{$bundle}";

    if ($cache = $this->cache->get($cid)) {
      return $cache->data;
    }

    $fields = [];
    $cache_tags = $this->entityTypeManager->getDefinition('meta_entity_type')->getListCacheTags();
    $storage = $this->entityTypeManager->getStorage('meta_entity_type');
    foreach ($storage->loadMultiple() as $id => $type) {
      $cache_tags = Cache::mergeTags($cache_tags, $type->getCacheTags());
      $mapping = $type->get('mapping');
      if (!empty($mapping[$entity_type_id][$bundle])) {
        $fields[$id] = $mapping[$entity_type_id][$bundle];
      }
    }

    if ($fields) {
      // Make returned list predictable.
      ksort($fields);
    }

    // Cache the results.
    $this->cache->set($cid, $fields, Cache::PERMANENT, $cache_tags);

    return $fields;
  }

}
